/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is DogStar SOFTWARE PiXML Public Module.
 *
 * The Initial Developer of the Original Code is
 * James Tyson, DogStar SOFTWARE <james@thedogstar.org>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   None.
 *
 * ***** END LICENSE BLOCK ***** */

constant XML = 1;
constant NODE = 2;

static float  _xml_version  = 1.0;
static string _xml_encoding = "ISO-8859-1";
static string _pike_xml_ns  = "http://pike.ida.liu.se/";
static string _pike_xml_xmlns  = "pike";
static string _string_encoding = "cdata";
static string _mime_xmlns_ns = "http://pike.ida.liu.se/generated/manual/modref/ex/predef_3A_3A/MIME.html";
static string _mime_xmlns_prefix = "mime";

//! Change the XML version number in the xml PI. Defaults to 1.0.
//! 
//! @param var
//!   Version number. Really should only be 1.0.
float xml_version(void|float ver) {
  if (ver && floatp(ver) && ver != _xml_version)
    _xml_version = ver;
  return _xml_version;
}

//! Change the XML encoding in the xml PI.
//! Defaults to "ISO-8859-1".
//! 
//! @param encoding
//!   The encoding.
string xml_encoding(void|string encoding) {
  if (encoding && stringp(encoding) && encoding != _xml_encoding)
    _xml_encoding = encoding;
  return _xml_encoding;
}

//! Change the default Pike namespace - don't do this unless you know
//! what you're doing.
//! Defaults to "http://pike.ida.liu.se/".
//!
//! @param ns
//!   The URI of the new namespace.
string pike_xml_namespace(void|string ns) {
  if (ns && stringp(ns) && sizeof(ns))
    _pike_xml_ns = ns;
  return _pike_xml_ns;
}

//! Change the default prefix for the Pike namespace - don't do this 
//! unless you know what you're doing.
//! Defaults to "pike".
//!
//! @param prefix
//!   The URI of the new namespace.
string pike_xml_ns_prefix(void|string prefix) {
  if (prefix && stringp(prefix) && sizeof(prefix))
    _pike_xml_xmlns = prefix;
  return _pike_xml_xmlns;
}

//! Change the default MIME namespace - don't do this unless you know
//! what you're doing.
//! Defaults to "http://pike.ida.liu.se/generated/manual/modref/ex/predef_3A_3A/MIME.html".
//!
//! @param ns
//!   The URI of the new namespace.
string mime_xml_namespace(void|string ns) {
  if (ns && stringp(ns) && sizeof(ns))
    _mime_xmlns_ns = ns;
  return _mime_xmlns_ns;
}

//! Change the default prefix for the MIME namespace - don't do this 
//! unless you know what you're doing.
//! Defaults to "mime".
//!
//! @param prefix
//!   The URI of the new namespace.
string mime_xml_ns_prefix(void|string prefix) {
  if (prefix && stringp(prefix) && sizeof(prefix))
    _mime_xmlns_prefix = prefix;
  return _mime_xmlns_prefix;
}

//! Change the default string encoding.  Defaults to "cdata".
//! Use this if you're trying to encode a binary string to
//! XML and you're getting backtraces.
//!
//! @param encoding
//!   Either "base64", "uue" or "cdata".
string string_encoding(void|string encoding) {
  if (encoding && stringp(encoding) && 
     (encoding == "base64" || encoding == "uue" || encoding == "cdata"))
    _string_encoding = encoding;
  return _string_encoding;
}

//! Serialise the argument into XML.
//! 
//! @param data
//!   The data to be encoded.
//!
//! @seealso PiJAX.supported()
string|object serialise(mixed data, void|int as) {
  if (!as)
    return low_serialise(data)->render_xml(1,1);
  else if (as == XML)
    return low_serialise(data)->render_xml(1,1);
  else if (as == NODE)
    return low_serialise(data);
}

static object low_serialise(mixed data, void|object node) {
  if (floatp(data))
    return floatx(data, node);
  else if (intp(data))
    return intx(data, node);
  else if (stringp(data))
    return stringx(data, node);
  else if (multisetp(data))
    return multisetx(data, node);
  else if (arrayp(data))
    return arrayx(data, node);
  else if (mappingp(data))
    return mappingx(data, node);
#if constant(Image.PNG.encode)
  else if (sprintf("%t", data) == "Image.Image")
    return imagex(data, node);
#endif
  else
    return exceptionx(sprintf("Error.  Don't know how to serialise %O\n", data), node);
}

static object new_elem(string node_name, void|object parent) {
  if (!parent) {
    parent = Public.Parser.XML2.new_xml("1.0", node_name);
    parent->add_ns(pike_xml_namespace(), pike_xml_ns_prefix());
    parent->add_ns(mime_xml_namespace(), mime_xml_ns_prefix());
    parent->set_ns(pike_xml_namespace());
    object pi = parent->new_pi("xml", sprintf("version=\"%:02f\" encoding=\"%s\"", xml_version(), xml_encoding()));
    parent->add_sibling(pi);
    return parent;
  }
  else {
    object node = parent->add_child(Public.Parser.XML2.new_node(node_name));
    node->set_ns(pike_xml_namespace());
    return node;
  }
}

static object floatx(float data, void|object node) {
  node = new_elem("float", node);
  node->set_attribute("value", pike_xml_namespace(), sprintf("%f", data));
  return node;
}

static object intx(int data, void|object node) {
  node = new_elem("integer", node);
  node->set_attribute("value", pike_xml_namespace(), sprintf("%d", data));
  return node;
}

static object stringx(string data, void|object node) {
  node = new_elem("string", node);
  object cdata;
  switch(string_encoding()) {
    case "uue":
      node->set_attribute("encoding", mime_xml_namespace(), "base64");
      cdata = node->new_cdata_block(MIME.encode_uue(data));
      node->add_child(cdata);
      break;
    case "base64":
      node->set_attribute("encoding", mime_xml_namespace(), "base64");
      cdata = node->new_cdata_block(MIME.encode_base64(data));
      node->add_child(cdata);
      break;
    default:
      cdata = node->new_cdata_block(data);
      node->add_child(cdata);
      break;
  }
  return node;
}

static object multisetx(multiset data, void|object node) {
  node = new_elem("multiset", node);
  node->set_attribute("elements", pike_xml_namespace(), (string)sizeof(data));
  foreach(sort(indices(data)), mixed idx) {
    object element = new_elem("element", node);
    low_serialise(idx, element);
  }
  return node;
}

static object arrayx(array data, void|object node) {
  node = new_elem("array", node);
  node->set_attribute("elements", pike_xml_namespace(), (string)sizeof(data));
  for(int i = 0; i < sizeof(data); i++) {
    object element = new_elem("element", node);
    element->set_attribute("index", pike_xml_namespace(), (string)i);
    low_serialise(data[i], element);
  }
  return node;
}

static object mappingx(mapping data, void|object node) {
  node = new_elem("mapping", node);
  node->set_attribute("elements", pike_xml_namespace(), (string)sizeof(data));
  foreach(sort(indices(data)), mixed idx) {
    object map = new_elem("map", node);
    object index = new_elem("index", map);
    object element = new_elem("element", map);
    low_serialise(idx, index);
    low_serialise(data[idx], element);
  }
  return node;
}

#if constant(Image.PNG.encode)
static object imagex(object data, void|object node) {
  node = new_elem("image", node);
  node->set_attribute("type", mime_xml_namespace(), "image/png");
  node->set_attribute("encoding", mime_xml_namespace(), "base64");
  object cdata = node->new_cdata_block(MIME.encode_base64(Image.PNG.encode(data)));
  node->add_child(cdata);
  return node;
}
#endif

static object exceptionx(string data, void|object node) {
  node = new_elem("exception", node);
  object cdata = node->new_cdata_block(data);
  node->add_child(cdata);
  return node;
}
